@@ -1,20 +1,20 @@
 # Copyright 2013 The py-lmdb authors, all rights reserved.
-# 
+#
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted only as authorized by the OpenLDAP
 # Public License.
-# 
+#
 # A copy of this license is available in the file LICENSE in the
 # top-level directory of the distribution or, alternatively, at
 # <http://www.OpenLDAP.org/license.html>.
-# 
+#
 # OpenLDAP is a registered trademark of the OpenLDAP Foundation.
-# 
+#
 # Individual files and/or contributed packages may be copyright by
 # other parties and/or subject to additional restrictions.
-# 
+#
 # This work also contains materials derived from public sources.
-# 
+#
 # Additional information about OpenLDAP can be obtained at
 # <http://www.openldap.org/>.
 
@@ -27,20 +27,22 @@ Please see http://lmdb.readthedocs.org/
 import os
 import sys
 
-try:
+def _reading_docs():
     # Hack: disable speedups while testing or reading docstrings.
-    if os.path.basename(sys.argv[0]) in ('sphinx-build', 'pydoc') or \
-            os.getenv('LMDB_FORCE_CFFI') is not None:
+    basename = os.path.basename(sys.argv[0])
+    return any(x in basename for x in ('sphinx-build', 'pydoc'))
+
+try:
+    if _reading_docs() or os.getenv('LMDB_FORCE_CFFI') is not None:
         raise ImportError
     from lmdb.cpython import *
 except ImportError:
     from lmdb.cffi import *
     from lmdb.cffi import __doc__
 
-del os
 __all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
            'enable_drop_gil']
-__version__ = '0.69'
+__version__ = '0.70'
 
 # Hack to support Python v2.5 'python -mlmdb'
 if __name__ == '__main__':
