@@ -19,7 +19,7 @@
 # <http://www.openldap.org/>.
 
 """
-cffi wrapper for OpenLDAP's "Lightning" MDB database.
+CPython/cffi wrapper for OpenLDAP's "Lightning" MDB database.
 
 Please see http://lmdb.readthedocs.org/
 """
@@ -33,7 +33,7 @@ import threading
 import warnings
 import weakref
 
-import cffi
+import lmdb
 
 __all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
            'enable_drop_gil']
@@ -41,8 +41,7 @@ __all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
 # Used to track context across cffi callbcks.
 _callbacks = threading.local()
 
-_ffi = cffi.FFI()
-_ffi.cdef('''
+_CFFI_CDEF = '''
     typedef int mode_t;
     typedef ... MDB_env;
     typedef struct MDB_txn MDB_txn;
@@ -179,9 +178,9 @@ _ffi.cdef('''
     static int pymdb_cursor_put(MDB_cursor *cursor,
                                 char *key_s, size_t keylen,
                                 char *val_s, size_t vallen, int flags);
-''')
+'''
 
-_lib = _ffi.verify('''
+_CFFI_VERIFY = '''
     #include <sys/stat.h>
     #include "lmdb.h"
 
@@ -234,16 +233,27 @@ _lib = _ffi.verify('''
         MDB_val tmpval = {vallen, val_s};
         return mdb_cursor_put(cursor, &tmpkey, &tmpval, flags);
     }
-''',
-    ext_package='lmdb',
-    sources=['lib/mdb.c', 'lib/midl.c'],
-    extra_compile_args=['-Wno-shorten-64-to-32'],
-    include_dirs=['lib']
-)
+'''
 
-globals().update((k, getattr(_lib, k))
-                 for k in dir(_lib) if k[:4] in ('mdb_', 'MDB_', 'pymd'))
-EINVAL = _lib.EINVAL
+if not lmdb._reading_docs():
+    import cffi
+    _ffi = cffi.FFI()
+    _ffi.cdef(_CFFI_CDEF)
+    _lib = _ffi.verify(_CFFI_VERIFY,
+        ext_package='lmdb',
+        sources=['lib/mdb.c', 'lib/midl.c'],
+        extra_compile_args=['-Wno-shorten-64-to-32'],
+        include_dirs=['lib'])
+    globals().update((k, getattr(_lib, k))
+                     for k in dir(_lib) if k[:4] in ('mdb_', 'MDB_', 'pymd'))
+    EINVAL = _lib.EINVAL
+
+    @_ffi.callback("int(char *, void *)")
+    def _msg_func(s, _):
+        """mdb_msg_func() callback. Appends `s` to _callbacks.msg_func list.
+        """
+        _callbacks.msg_func.append(_ffi.string(s))
+        return 0
 
 class Error(Exception):
     """Raised when any MDB error occurs."""
@@ -326,13 +336,6 @@ def enable_drop_gil():
     *Caution:* this function should be invoked before any threads are created.
     """
 
-@_ffi.callback("int(char *, void *)")
-def _msg_func(s, _):
-    """mdb_msg_func() callback. Appends `s` to _callbacks.msg_func list.
-    """
-    _callbacks.msg_func.append(_ffi.string(s))
-    return 0
-
 
 class Environment(object):
     """
@@ -359,9 +362,11 @@ class Environment(object):
             On 64-bit there is no penalty for making this huge (say 1TB). Must
             be <2GB on 32-bit.
 
-            *Note*: **the default map size is set low to encourage a crash**,
-            so users can figure out a good value before learning about this
-            option too late.
+            .. note::
+
+                **The default map size is set low to encourage a crash**, so
+                users can figure out a good value before learning about this
+                option too late.
 
         `subdir`:
             If ``True``, `path` refers to a subdirectory to store the data and
@@ -391,8 +396,13 @@ class Environment(object):
 
         `writemap`:
             If ``True`` LMDB will use a writeable memory map to update the
-            database. Requires operating system support. Please refer to
-            documentation.
+            database. This option is incompatible with nested transactions.
+
+        'map_async':
+             When ``writemap=True``, use asynchronous flushes to disk. As with
+             ``sync=False``, a system crash can then corrupt the database or
+             lose the last transactions. Calling :py:meth:`sync` ensures
+             on-disk database integrity until next commit.
 
         `max_readers`:
             Maximum number of simultaneous read transactions. Can only be set
@@ -470,6 +480,9 @@ class Environment(object):
         if writemap:
             flags |= MDB_WRITEMAP
 
+        if isinstance(path, type(u'')):
+            path = path.encode(sys.getfilesystemencoding())
+
         rc = mdb_env_open(self._env, path, flags, mode)
         if rc:
             raise Error(path, rc)
