@@ -20,6 +20,8 @@
  * <http://www.openldap.org/>.
  */
 
+#define PY_SSIZE_T_CLEAN
+
 #include <errno.h>
 #include <stdbool.h>
 #include <stdint.h>
@@ -30,6 +32,10 @@
 #include "Python.h"
 #include "structmember.h"
 
+#ifdef HAVE_MEMSINK
+#include "memsink.h"
+#endif
+
 #include "lmdb.h"
 
 
@@ -166,13 +172,13 @@ static PyObject *Error;
 /** If 1, save_thread() and restore_thread() drop GIL. */
 static int drop_gil = 0;
 
-extern PyTypeObject PyDatabase_Type;
-extern PyTypeObject PyEnvironment_Type;
-extern PyTypeObject PyTransaction_Type;
-extern PyTypeObject PyCursor_Type;
-extern PyTypeObject PyIterator_Type;
-
 /** Typedefs and forward declarations. */
+static PyTypeObject PyDatabase_Type;
+static PyTypeObject PyEnvironment_Type;
+static PyTypeObject PyTransaction_Type;
+static PyTypeObject PyCursor_Type;
+static PyTypeObject PyIterator_Type;
+
 typedef struct CursorObject CursorObject;
 typedef struct DbObject DbObject;
 typedef struct EnvObject EnvObject;
@@ -305,6 +311,10 @@ enum trans_flags {
 struct TransObject {
     LmdbObject_HEAD
     EnvObject *env;
+#ifdef HAVE_MEMSINK
+    /** Copy-on-invalid list head. */
+    PyObject *sink_head;
+#endif
     /** MDB transaction object. */
     MDB_txn *txn;
     /** Bitfield of trans_flags values. */
@@ -991,6 +1001,9 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
     Py_INCREF(env);
     self->db = db;
     Py_INCREF(db);
+#ifdef HAVE_MEMSINK
+    self->sink_head = NULL;
+#endif
     self->key_buf = NULL;
 
     self->flags = 0;
@@ -1115,7 +1128,7 @@ db_dealloc(DbObject *self)
     PyObject_Del(self);
 }
 
-PyTypeObject PyDatabase_Type = {
+static PyTypeObject PyDatabase_Type = {
     PyObject_HEAD_INIT(NULL)
     .tp_basicsize = sizeof(DbObject),
     .tp_dealloc = (destructor) db_dealloc,
@@ -1890,7 +1903,7 @@ static struct PyMethodDef env_methods[] = {
 };
 
 
-PyTypeObject PyEnvironment_Type = {
+static PyTypeObject PyEnvironment_Type = {
     PyObject_HEAD_INIT(0)
     .tp_basicsize = sizeof(EnvObject),
     .tp_dealloc = (destructor) env_dealloc,
@@ -2396,7 +2409,7 @@ static struct PyMethodDef cursor_methods[] = {
     {NULL, NULL}
 };
 
-PyTypeObject PyCursor_Type = {
+static PyTypeObject PyCursor_Type = {
     PyObject_HEAD_INIT(0)
     .tp_basicsize = sizeof(CursorObject),
     .tp_dealloc = (destructor) cursor_dealloc,
@@ -2456,7 +2469,7 @@ static struct PyMethodDef iter_methods[] = {
     {NULL, NULL}
 };
 
-PyTypeObject PyIterator_Type = {
+static PyTypeObject PyIterator_Type = {
     PyObject_HEAD_INIT(0)
     .tp_basicsize = sizeof(IterObject),
     .tp_dealloc = (destructor) iter_dealloc,
@@ -2477,6 +2490,9 @@ trans_clear(TransObject *self)
 {
     if(self->valid) {
         INVALIDATE(self)
+#ifdef HAVE_MEMSINK
+        ms_notify((PyObject *) self, &self->sink_head);
+#endif
         if(self->txn) {
             DEBUG("aborting")
             DROP_GIL
@@ -2539,6 +2555,9 @@ trans_abort(TransObject *self)
     }
     DEBUG("aborting")
     INVALIDATE(self)
+#ifdef HAVE_MEMSINK
+    ms_notify((PyObject *) self, &self->sink_head);
+#endif
     DROP_GIL
     mdb_txn_abort(self->txn);
     LOCK_GIL
@@ -2555,6 +2574,9 @@ trans_commit(TransObject *self)
     }
     DEBUG("committing")
     INVALIDATE(self)
+#ifdef HAVE_MEMSINK
+    ms_notify((PyObject *) self, &self->sink_head);
+#endif
     int rc;
     UNLOCKED(rc, mdb_txn_commit(self->txn));
     self->txn = NULL;
@@ -2694,7 +2716,7 @@ static struct PyMethodDef trans_methods[] = {
     {NULL, NULL}
 };
 
-PyTypeObject PyTransaction_Type = {
+static PyTypeObject PyTransaction_Type = {
     PyObject_HEAD_INIT(0)
     .tp_basicsize = sizeof(TransObject),
     .tp_dealloc = (destructor) trans_dealloc,
@@ -2818,6 +2840,14 @@ MODINIT_NAME(void)
     if(init_types(mod)) {
         MOD_RETURN(NULL);
     }
+
+#ifdef HAVE_MEMSINK
+    MemSink_IMPORT;
+    if(ms_init_source(&PyTransaction_Type, offsetof(TransObject, sink_head))) {
+        MOD_RETURN(NULL);
+    }
+#endif
+
     if(init_strings(mod)) {
         MOD_RETURN(NULL);
     }
