@@ -34,16 +34,31 @@ Basic tools for working with LMDB.
         If no databases are given, rewrites only the main database.
 
     shell: Open interactive console with ENV set to the open environment.
+
+    stat: Print environment statistics.
+
+    warm: Read environment into page cache sequentially.
+
+    watch: Show live environment statistics
 """
 
 from __future__ import absolute_import
 from __future__ import with_statement
+import array
+import collections
 import contextlib
+import csv
+import fcntl
 import functools
 import optparse
 import os
+import pprint
+import signal
 import string
+import struct
 import sys
+import termios
+import time
 
 # Python3.x bikeshedded trechery.
 try:
@@ -129,7 +144,13 @@ def make_parser():
                      help='List of key pairs to read from files.')
     group.add_option('--delete', action='append',
                      help='List of key=value pairs to delete.')
-
+    group = parser.add_option_group('Options for "watch" command')
+    group.add_option('--csv', action='store_true',
+                     help='Generate CSV instead of terminal output.')
+    group.add_option('--interval', type='int', default=1,
+                     help='Interval size (default: 1sec)')
+    group.add_option('--window', type='int', default=10,
+                     help='Average window size (default: 10)')
     return parser
 
 
@@ -267,6 +288,96 @@ def cmd_restore(opts, args):
                 print('Loaded %d keys from %r' % (count, path))
 
 
+def delta(hst):
+    return [(hst[i] - hst[i-1]) for i in xrange(1, len(hst))]
+
+def cmd_watch(opts, args):
+    info = None
+    stat = None
+
+    def window(func):
+        history = collections.deque()
+        def windowfunc():
+            history.append(func())
+            if len(history) > opts.window:
+                history.popleft()
+            if len(history) <= 1:
+                return 0
+            n = sum(delta(history)) / float(len(history) - 1)
+            return n / opts.interval
+        return windowfunc
+
+    envmb = lambda: (info['last_pgno'] * stat['psize']) / 1048576.
+
+    cols = [
+        ('%d',    'Depth', lambda: stat['depth']),
+        ('%d',    'Branch', lambda: stat['branch_pages']),
+        ('%d',    'Leaf', lambda: stat['leaf_pages']),
+        ('%+d',   'Leaf/s', window(lambda: stat['leaf_pages'])),
+        ('%d',    'Oflow', lambda: stat['overflow_pages']),
+        ('%+d',   'Oflow/s', window(lambda: stat['overflow_pages'])),
+        ('%d',    'Recs', lambda: stat['entries']),
+        ('%+d',   'Recs/s', window(lambda: stat['entries'])),
+        ('%d',    'Rdrs', lambda: info['num_readers']),
+        ('%.2f',  'EnvMb', envmb),
+        ('%+.2f', 'EnvMb/s', window(envmb)),
+        ('%d',    'Txs', lambda: info['last_txnid']),
+        ('%+.2f', 'Txs/s', window(lambda: info['last_txnid']))
+    ]
+
+    term_width = 0
+    widths = [len(head) for _, head, _ in cols]
+
+    if opts.csv:
+        writer = csv.writer(sys.stdout, quoting=csv.QUOTE_ALL)
+        writer.writerow([head for _, head, _ in cols])
+
+    cnt = 0
+    try:
+        while True:
+            stat = ENV.stat()
+            info = ENV.info()
+
+            vals = []
+            for i, (fmt, head, func) in enumerate(cols):
+                val = fmt % func()
+                vals.append(val)
+                widths[i] = max(widths[i], len(val))
+
+            if opts.csv:
+                writer.writerow(vals)
+            else:
+                if term_width != _TERM_WIDTH or not (cnt % (_TERM_HEIGHT - 2)):
+                    for i, (fmt, head, func) in enumerate(cols):
+                        sys.stdout.write(head.rjust(widths[i] + 1))
+                    sys.stdout.write('\n')
+                    term_width = _TERM_WIDTH
+                for i, val in enumerate(vals):
+                    sys.stdout.write(val.rjust(widths[i] + 1))
+                sys.stdout.write('\n')
+
+            time.sleep(opts.interval)
+            cnt += 1
+    except KeyboardInterrupt:
+        pass
+
+
+def cmd_warm(opts, args):
+    stat = ENV.stat()
+    info = ENV.info()
+
+    bufsize = 32768
+    last_offset = stat['psize'] * info['last_pgno']
+    buf = array.array('c', '\x00' * bufsize)
+    t0 = time.time()
+
+    fp = open(opts.env + '/data.mdb', 'rb', bufsize)
+    while fp.tell() < last_offset:
+        fp.readinto(buf)
+    print('Warmed %.2fmb in %dms' %
+        (last_offset / 1048576., 1000 * (time.time() - t0)))
+
+
 def cmd_rewrite(opts, args):
     if not opts.target_env:
         die('Must specify target environment path with -E')
@@ -274,8 +385,9 @@ def cmd_rewrite(opts, args):
     src_info = ENV.info()
     target_env = lmdb.open(opts.target_env,
                            map_size=src_info['map_size'] * 2,
-                           max_dbs=opts.max_dbs,
-                           writemap=True)
+                           max_dbs=opts.max_dbs, sync=False,
+                           writemap=True, map_async=True,
+                           metasync=False)
 
     dbs = []
     for arg in args:
@@ -293,6 +405,9 @@ def cmd_rewrite(opts, args):
             for key, value in ENV.cursor(db=src_db, buffers=True):
                 wtxn.put(key, value, append=True)
 
+    print('Syncing..')
+    target_env.sync(True)
+
 
 def cmd_get(opts, args):
     print_header = len(args) > 1
@@ -345,6 +460,23 @@ def cmd_shell(opts, args):
     code.InteractiveConsole(globals()).interact()
 
 
+def cmd_stat(opts, args):
+    pprint.pprint(ENV.stat())
+    pprint.pprint(ENV.info())
+
+
+def _get_term_width(default=80):
+    try:
+        s = fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, '1234')
+        height, width = struct.unpack('hh', s)
+        return width, height
+    except:
+        return default
+
+def _on_sigwinch(*args):
+    global _TERM_WIDTH, _TERM_HEIGHT
+    _TERM_WIDTH, _TERM_HEIGHT = _get_term_width()
+
 def main():
     parser = make_parser()
     opts, args = parser.parse_args()
@@ -362,6 +494,9 @@ def main():
         global DB
         DB = ENV.open_db(opts.db)
 
+    signal.signal(signal.SIGWINCH, _on_sigwinch)
+    _on_sigwinch()
+
     func = globals().get('cmd_' + args[0])
     if not func:
         die('No such command: %r' % (args[0],))
